package com.blackledge.david.tivo;

import java.awt.Rectangle;
import java.io.UnsupportedEncodingException;
import java.lang.ref.WeakReference;
import java.net.URLDecoder;
import java.util.Map;
import java.util.LinkedHashMap;

import com.tivo.hme.bananas.BApplication;
import com.tivo.hme.interfaces.IContext;
import com.tivo.hme.sdk.Application;
import com.tivo.hme.sdk.HmeEvent;
import com.tivo.hme.sdk.HmeEvent.DeviceInfo;
import com.tivo.hme.sdk.IHmeProtocol;
import com.tivo.hme.sdk.Resource;
import com.tivo.hme.sdk.Version;

/**
 * BApplication that fixes SDK bugs, supports some features beyond SDK 1.4.1 and
 * makes some things easier. See {@link V49BApplicationFactory} and
 * {@link V49IHmeProtocol} for other features.
 * <UL>
 * <LI>Use {@link V49BApplicationFactory} as superclass for your factory to tell the TiVo you're using V49 (or another version).
 * <LI>Constants for SDK resource strings *.ttf, *.snd
 * <LI>Constants for discovered Stream resources "livetv:" "loopset:" and
 * "recording:"
 * <LI>Getters for spacing when a "loopset:" is the background
 * <LI>Member variables for protocol version, tivo version, and tivo ID.
 * <LI>Resolution help: automatically set preferred resolution, template methods
 * for init and showing the first BScreen during resolution setting
 * <LI>Template method for cleaning up during {@link #destroy()} to prevent an
 * SDK 1.4 bug (not an issue in 1.4.1e)
 * <LI>Post Resource errors to the resource and with new version of
 * {@link #handleApplicationError(int, String, Integer)}
 * <LI>Fix SDK bug that broke Application-as-stream feature.
 * <LI>Support more TTF built-in fonts with new constant names for series-4-only
 * fonts, and with id numbers-as-font-names for arbitrary attempts.
 * <LI>SDK capability/version methods.
 * <LI>Backward compatibility of new SDK methods to old SDK. You should be able
 * to run this app in either SDK.
 * </UL>
 * 
 * @author David.Blackledge.com
 * 
 */
public abstract class V49BApplication extends BApplication implements V49IHmeProtocol {

    public static final String TTF_SERIES4FONT3BOLD = "series4font3bold.ttf";
	public static final String TTF_SERIES4FONTITALIC = "series4fontitalic.ttf";
	public static final String TTF_SERIES4FONT = "series4font.ttf";
	public static final String TTF_DEFAULT = "default.ttf"; //$NON-NLS-1$
    public static final String TTF_SYSTEM = "system.ttf"; //$NON-NLS-1$
    
    public static final String SOUND_BONK = "bonk.snd"; //$NON-NLS-1$
    public static final String SOUND_UPDOWN = "updown.snd"; //$NON-NLS-1$
    public static final String SOUND_THUMBSUP = "thumbsup.snd"; //$NON-NLS-1$
    public static final String SOUND_THUMBSDOWN = "thumbsdown.snd"; //$NON-NLS-1$
    public static final String SOUND_SELECT = "select.snd"; //$NON-NLS-1$
    public static final String SOUND_TIVO = "tivo.snd"; //$NON-NLS-1$
    public static final String SOUND_LEFT = "left.snd"; //$NON-NLS-1$
    public static final String SOUND_RIGHT = "right.snd"; //$NON-NLS-1$
    public static final String SOUND_PAGEUP = "pageup.snd"; //$NON-NLS-1$
    public static final String SOUND_PAGEDOWN = "pagedown.snd"; //$NON-NLS-1$
    public static final String SOUND_ALERT = "alert.snd"; //$NON-NLS-1$
    public static final String SOUND_DESELECT = "deselect.snd"; //$NON-NLS-1$
    public static final String SOUND_ERROR = "error.snd"; //$NON-NLS-1$
    public static final String SOUND_SLOWDOWN1 = "slowdown1.snd"; //$NON-NLS-1$
    public static final String SOUND_SPEEDUP1 = "speedup1.snd"; //$NON-NLS-1$
    public static final String SOUND_SPEEDUP2 = "speedup2.snd"; //$NON-NLS-1$
    public static final String SOUND_SPEEDUP3 = "speedup3.snd"; //$NON-NLS-1$
    
	public static final String STREAM_LIVETV = "livetv:"; //$NON-NLS-1$
	public static final String STREAM_LOOP_RED_CENTRAL = "loopset:Central"; //$NON-NLS-1$
	public static final String STREAM_LOOP_PURPLE_SHOWCASES = "loopset:Showcases"; //$NON-NLS-1$
	public static final String STREAM_LOOP_BLUE_SETUP = "loopset:Setup"; //$NON-NLS-1$
	public static final String STREAM_LOOP_GREEN_NOWPLAYING = "loopset:NowPlaying"; //$NON-NLS-1$
	public static final String STREAM_RECORDING_PREFIX = "recording:"; //$NON-NLS-1$
	
    
	/**
	 * Protocol in use between the TiVo and the server. Used in methods like
	 * {@link #isSlideRemoteSupported()}. Accurate/dynamic version of
	 * {@link IHmeProtocol#VERSION}. Saved during
	 * {@link #setContext(IContext, int)} call. Use V49BapplicationFactory as
	 * your factory's supertype to register V49 as your preferred protocol
	 * version.
	 */
	public int protocolVersion;
	/**
	 * Device Info event's "version" - the TiVo's Software Version. E.g.
	 * 11.0k-01-2-652 . Available after DEVICE_INFO_EVENT comes in.
	 */
	protected String tivoVersion;
	/**
	 * The TiVo Service Number (tsn) uniquely identifying the connected TiVo DVR
	 * device. Available after {@link #setContext(IContext, int)} is called.
	 */
	protected String tivoID;

	/**
	 * Does nothing.  Override this to output debug messages
	 * @param message
	 */
	public void debug(String message) {
	}

	/**
	 * Template method to do {@link #init(IContext)} tasks between super's init
	 * and preferred resolution setting. Better to override this than init
	 * unless you want to handle setting resolution yourself.
	 */
	protected void initBeforeResolutionSet(IContext context) {
	}

	/**
	 * Template method to do {@link #rootBoundsChanged(Rectangle)} tasks between
	 * super's rootBoundsChanged and a possible call to {@link #showFirstPage()}
	 * if this is the first root bounds change during application
	 * initialization. Override this, or it might be OK to override
	 * rootBoundsChanged and just call its super method and ignore this.
	 * 
	 * @param rootBounds
	 */
	protected void rootBoundsChangedBeforeFirstPage(Rectangle rootBounds) {
	}

	/**
	 * Override this to push the first BScreen of the application after
	 * resolution issues are resolved.
	 */
	protected abstract void showFirstPage();

	/**
	 * Override this instead of {@link #destroy()} for SDK 1.4 to do safe cleanup during the
	 * destroy operation. Prevents the loop that occurs if a fatal error happens
	 * during destroy.  Not an issue in 1.4.1e
	 */
	protected void destroyCleanup() {
	}
	
	
	/* BEGIN fix JDK */
	

	// TODO future: handle unsupported chunks (e.g. Event 10)
	// public boolean handleChunk(InputStream in)
	// {
	// synchronized (lock) {
	// //debug("handleChunk()");
	// boolean doMore = true;
	//
	// // flush any data that was generated from previous handling of
	// // event
	// flush();
	//
	// ChunkedInputStream chunkInStr = null;
	// if (in instanceof ChunkedInputStream)
	// {
	// chunkInStr = (ChunkedInputStream)in;
	// }
	// else
	// {
	// chunkInStr = new ChunkedInputStream(in); //,
	// IHmeConstants.TCP_BUFFER_SIZE);
	// }
	//
	// // if ( protocolVersion < VERSION_0_40 )
	// // {
	// // chunkInStr.setUseVString( false );
	// // }
	// // else {
	// chunkInStr.setUseVString( true );
	// // }
	//
	// int opcode = -1;
	// try {
	// opcode = (int)chunkInStr.readVInt();
	// } catch (IOException e) {
	// // receiver closed - ignore
	// // if (debug()) {
	// // log(ILogger.LOG_DEBUG,
	// // "Connection terminated by receiver");
	// // }
	// }
	// if (opcode == -1) {
	// doMore = false;
	// return doMore;
	// }
	//
	// HmeEvent evt = null;
	// String eventString = null;
	// try
	// {
	// switch (opcode) {
	// case EVT_DEVICE_INFO:
	// evt = new HmeEvent.DeviceInfo(chunkInStr);
	// break;
	// case EVT_APP_INFO:
	// evt = new HmeEvent.ApplicationInfo(chunkInStr);
	// break;
	// case EVT_RSRC_INFO:
	// evt = new HmeEvent.ResourceInfo(chunkInStr, this);
	// break;
	// case EVT_RESOLUTION_INFO:
	// evt = new HmeEvent.ResolutionInfo(chunkInStr);
	// break;
	// case EVT_KEY:
	// evt = new HmeEvent.Key(chunkInStr);
	// break;
	// case EVT_IDLE:
	// evt = new HmeEvent.Idle(chunkInStr);
	// break;
	// case EVT_FONT_INFO:
	// evt = new HmeEvent.FontInfo(chunkInStr);
	// break;
	// case EVT_INIT_INFO:
	// evt = new HmeEvent.InitInfo(chunkInStr);
	// break;
	// default:
	// eventString =
	// "HmeEvent[opcode=" + opcode
	// + ", id = " + chunkInStr.readVInt()
	// + "]";
	// // short clen = chunkInStr.readShort();
	// // while (clen != 0) {
	// // System.out.println(clen);
	// // chunkInStr.skip(clen);
	// // clen = chunkInStr.readShort();
	// // }
	//
	// // System.out.println(eventString);
	// // try {
	// // System.out.println(chunkInStr.readInt());
	// // System.out.println(chunkInStr.readDict());
	// // } catch(IOException e1) {
	// // chunkInStr.readTerminator();
	// // throw e1;
	// // }
	// break;
	// }
	// if (evt != null) {
	// evt.setMonitor(lock);
	// }
	// chunkInStr.readTerminator();
	// }
	// catch (IOException e) {
	// evt = null;
	// getLogger().log(ILogger.LOG_INFO, e);
	// }
	//
	// //
	// // Save the event, in case we need to dump the list
	// //
	// // if (eventsReceived != null) {
	// // if (eventsReceived.size() >= MAX_RECEIVED_EVENTS_SIZE) {
	// // // Startup is complete -- stop checking for protocol errs
	// // startupComplete = true;
	// // eventsReceived.clear();
	// // eventsReceived = null;
	// // } else {
	// // if (evt != null) {
	// // eventString = evt.toString();
	// // }
	// // if (eventString == null) {
	// // eventString = "Bad HmeEvent[opcode="+opcode+", no ID]";
	// // }
	// // eventsReceived.add(eventString);
	// // }
	// // }
	//
	// if (evt == null) {
	// getLogger().log(ILogger.LOG_DEBUG,
	// "unknown event opcode : " + opcode);
	// }
	// else
	// {
	// // if any debugging is turned on, verify the event is kosher
	// // if (debug()) {
	// // evt.verify(logger);
	// // }
	//
	// // debugEvents means to print each and every event received
	// // if (debugEvents()) {
	// getLogger().log(ILogger.LOG_DEBUG, "event " + evt);
	// // }
	//
	// // if (!preprocessEvent(evt)) {
	// //
	// // // preprocessEvent() says this event should be discarded
	// // if (eventsReceived != null) {
	// // int i = eventsReceived.size() - 1;
	// // eventsReceived.set(i,
	// // eventsReceived.get(i).toString() + " [IGNORED]");
	// // }
	//
	// // } else {
	//
	// // The app should receive this event.
	// //
	// // Only dispatch events after app.init() is called.
	// // Save events before init(), to be dispatched later.
	// //
	// // if (appInitialized) {
	// // dispatchEvent(evt);
	// // } else if (queuedEvents != null) {
	// // // Save the events to send after initialization
	// // queuedEvents.add(evt);
	// // }
	// // }
	// }
	//
	// // flush any data that was generated from handling of event
	// flush();
	//
	// return doMore;
	// } // sync
	//
	// } // handleChunk()


	/**
	 * unfortunately sdk's Resource.ResourceStatus isn't public, so we recreate
	 * it here - all it does is use a different Event ID
	 */
	protected static class MockResourceStatus extends HmeEvent.ResourceInfo {
		MockResourceStatus(int id, Resource rsrc, int status, Map map) {
			super(EVT_RSRC_STATUS, id, rsrc, status, map);
		}

		public String toString() {
			return getID() + ".RSRC_STATUS(" + statusToString(getStatus()) //$NON-NLS-1$
					+ ")"; //$NON-NLS-1$
		}
	}

	private boolean destroyedAlready = false;

	public void destroy() {
		// protect against SDK 1.4 bug: if there is a fatal error during below
		// operations (e.g. remove operations) close gets called which calls
		// destroy again.
	
		// what happens: after the 2nd call to here, we'll get HMEException from
		// the remove call that caused it which contains the actual cause
		// (unless it was originally an HMEException)
		// setId(-1) has not been called on the object yet
	
		// SO: want to catch HMEExceptions if we want to finish our destroy
		// operation
		if (destroyedAlready)
			return;
		destroyedAlready = true;
	
		// template method
		destroyCleanup();
		// ... operations that could cause a fatal error, e.g. "remove" calls
	
		super.destroy(); // in case e.g. BApplication ever starts doing
							// something there.
	}
	
	/**
	 * Improved version of
	 * {@link Application#handleApplicationError(int, String)} that includes the
	 * resourceId related to the error - critical for knowing when you have a
	 * problem with a resource you created. Default implementation simply calls
	 * {@link #handleApplicationError(int, String)}. ApplicationError codes 1
	 * (e.g. "can't create rsrc. unsupported stream type
	 * /avatar/cf1e61a4330e75d5d1d7a744c5ef38c4") and 3 (e.g. "resource 2091 not
	 * found (type type[-1])") about a resource never provide a resourceinfo
	 * event!
	 * 
	 * @param errorCode
	 *            the error code of the error (one of the {@link IHmeProtocol}
	 *            APP_ERROR_ constants)
	 * @param errorText
	 *            details of the error
	 * @param resourceId
	 *            the id associated with the error - could be null. Suitable for
	 *            use with {@link #getResource(Object)} (or ((WeakReference)
	 *            {@link #getResources()}.get(Object)).get() to avoid automatic
	 *            SDK warning message when it's not found)
	 * @return true if the event was handled.
	 */
	public boolean handleApplicationError(int errorCode, String errorText, Integer resourceId) {
		if (handleApplicationError(errorCode, errorText)) {
			return true;
		} else {
			// FIXME add processing needed by e.g. image exceeding max width or
			// height (errorCode 2) with no resource id
			return false;
		}
	}

	/**
	 * Perform some common handling - be sure to call this from subclasses that
	 * override it. Sets {@link #tivoVersion} from DEVICE_INFO. Improves error
	 * handling by posting resource errors to the resource itself, and calling
	 * {@link #handleApplicationError(int, String, Integer)}, passing in
	 * resource ID.
	 * 
	 */
	public boolean handleEvent(HmeEvent event) {
		// get tivo version from device info
		if (event.getOpCode() == EVT_DEVICE_INFO) {
			HmeEvent.DeviceInfo info = (DeviceInfo) event;
			this.tivoVersion = (String) info.getMap().get("version"); //$NON-NLS-1$
			// com.tivo.examine.idType.stationId=tivo,
			// platform=Gen06,
			// host=Bedroom,
			// brand=TiVo,
			// com.tivo.examine.showing=true,
			// com.tivo.examine.idType.contentId=tivo,
			// version=11.0k-01-2-652
		}
	
		// ...
		if (event.getOpCode() == EVT_APP_INFO) {
			HmeEvent.ApplicationInfo info = (HmeEvent.ApplicationInfo) event;
			if (info.getMap().get("error.code") != null) { //$NON-NLS-1$
				int errorCode = Integer.parseInt((String) info.getMap().get(
						"error.code")); //$NON-NLS-1$
				Integer resourceId = null;
				if (info.getMap().get("error.rsrc") != null) { //$NON-NLS-1$
					try {
						resourceId = new Integer((String) info.getMap().get(
								"error.rsrc")); //$NON-NLS-1$
					} catch (NumberFormatException e) {
						resourceId = null;
					}
				}
				String errorText = (String) info.getMap().get("error.text"); //$NON-NLS-1$
				if (resourceId != null) {
					WeakReference res_wr = (WeakReference) getResources().get(
							resourceId);
					if (res_wr != null && res_wr.get() != null) {
						Resource res = (Resource) res_wr.get();
						Map resInfoMap = info.getMap();
						// TODO we're doing this regardless of whether it's
						// actually a change in status
						if (res.status == Resource.RSRC_STATUS_ERROR) {
							debug("NOTE: had to send ResourceStatus event for error even though resource's status was already error."); //$NON-NLS-1$
						}
						debug("Posting APP_INFO resource error to resource " //$NON-NLS-1$
								+ resourceId + ": " + resInfoMap); //$NON-NLS-1$
						res.postEvent(new MockResourceStatus(resourceId
								.intValue(), res, Resource.RSRC_STATUS_ERROR,
								resInfoMap));
					}
				} else
					debug("APP_INFO error " + errorCode + ": " + errorText); //$NON-NLS-1$ //$NON-NLS-2$
	
				return handleApplicationError(errorCode, errorText, resourceId);
			}
		}
	
		boolean result = super.handleEvent(event);
		// ...
	
		return result;
	}

	/**
	 * SDK Bug Fix - OMG! sick bug in HMEObject that has never been noticed or
	 * exercised since nobody has apparently attempted to do App-in-app with
	 * {@link #createStream(String, Map)}. They forgot a "new" keyword and
	 * auto-"fixed" the problem creating a method named LinkedHashMap that has
	 * default method content returning null! Also, calling code won't handle
	 * the nulls that are returned in certain circumstances, so now returning an
	 * empty map here. Lucky for us it is public and non-final. {@inheritDoc}
	 * 
	 * @see com.tivo.hme.sdk.HmeObject#parseQuery(java.lang.String)
	 */
	public Map parseQuery(String query) {
		Map map = new LinkedHashMap();
		if (query == null || query.indexOf('=') == -1) {
			return map;
		}
	
		int at = 0;
		int len = query.length();
	
		if (query.startsWith("?")) { //$NON-NLS-1$
			++at;
		}
		do {
			// find = and &
			int equal = query.indexOf('=', at);
			if (equal < 0) {
				throw new IllegalArgumentException(
						"invalid query (trailing key, = not found)"); //$NON-NLS-1$
			}
			int amp = query.indexOf('&', equal);
			if (amp == -1) {
				amp = query.length();
			}
	
			// add the key/value
			try {
				String key = URLDecoder.decode(query.substring(at, equal),
						"UTF-8"); //$NON-NLS-1$
				String value = URLDecoder.decode(
						query.substring(equal + 1, amp), "UTF-8"); //$NON-NLS-1$
				map.put(key.toLowerCase(), value);
			} catch (UnsupportedEncodingException e) {
				e.printStackTrace();
			}
	
			// advance
			at = amp + 1;
		} while (at < len);
		return map;
	}

	/* END fix JDK */

	/**
	 * Override to handle additional system font ids as [id].ttf and additional
	 * names.
	 */
	public Resource createTrueType(String family) {
		// automatically handle 10.ttf - 19.ttf as ID refs
		if (family != null && family.matches("\\d+\\.ttf")) {
			try {
				int id = Integer.parseInt(family.substring(0, family.length()-4));
				if (id < 20 && id >= 10) {
					return new TrueTypeFalseResource(this, family, id);
				}
				if(id >= 20) {
					return getResource(new Integer(id));
				}
			} catch (Exception e) {
			}
		}
		// handle specific new names
		if (family.equals(TTF_SERIES4FONT)) {
			return new TrueTypeFalseResource(this, family, 13);
		} else if (family.equals(TTF_SERIES4FONTITALIC)) {
			return new TrueTypeFalseResource(this, family, 14);
		} else if (family.equals(TTF_SERIES4FONT3BOLD)) {
			return new TrueTypeFalseResource(this, family, 15);
//		} else if... {
		} else {
			// default behavior
			return super.createTrueType(family);
		}
	}

	/**
	 * Equivalent of TrueTypeResource except accessible to this class so we can
	 * instantiate additional names with a supplied id
	 */
	private static class TrueTypeFalseResource extends Resource {
		String name;

		TrueTypeFalseResource(Application app, String name, int id) {
			super(app, id);
			this.name = name;
			((V49BApplication)app).debug("Created font " + name + " with id " + id);
		}

		/**
		 * This is a thread-safe public entry point.
		 */
		protected void toString(StringBuffer buf) {
			synchronized (lock) {
				buf.append(",ttf=" + name);
			}
		}
	}
	
	
	private boolean waitingToInit;

	@Override
	protected void setContext(IContext context, int version) {
		super.setContext(context, version);
		// make accessible version of private Application class variable.
		this.protocolVersion = version;
		debug("Protocol Version per TiVo DVR: " + protocolVersion); //$NON-NLS-1$
	
		debug("attributes: " + context.getConnectionAttributes()); //$NON-NLS-1$
		// User-Agent: TvHttpClient -my tivo (and tsn=64...3E )
		// User-Agent: TmkHttpRequest/1.0 - from somebody on internet - old
		// version of tivo?
		// User-Agent=tivo.http/03/30/2003 - simulator
		// attributes: {Cookie=id=895df24b5038b953631128d3aa994863d51da2b1,
		// Host=192.168.0.20:7288, User-Agent=TvHttpClient, Connection=close,
		// tsn=64...3E}
		this.tivoID = context.getConnectionAttribute("tsn"); //$NON-NLS-1$
		// context.getConnectionAttribute(HTTP_HEADER_TIVO_ID);
		// tivoVersion is in DEVICE_INFO event under "version"
		// this.tivoVersion =
		// context.getConnectionAttribute(HTTP_HEADER_TIVO_VERSION);
	
	}

	/**
	 * You should override {@link #initBeforeResolutionSet(IContext)} to do your
	 * init work while this override handles resolution details after that is
	 * called.
	 */
	@Override
	public void init(IContext context) throws Exception {
		super.init(context);
		setActive(true);
		
		// template method for subclasses to perform init activities
		initBeforeResolutionSet(context);
		
		waitingToInit = true;
		boolean changed = setPreferredResolution();
		// will continue init on rootBoundsChanged()
	
		this.setActive(true);
	
		if (!changed || !isResolutionSupported()) {
			waitingToInit = false;
			showFirstPage();
		}

	}

	/**
	 * This override coordinates with the {@link #init(IContext)} override to
	 * call {@link #showFirstPage()} at the right time. Should probably override
	 * {@link #rootBoundsChangedBeforeFirstPage(Rectangle)} instead of this to
	 * maintain order of calls.
	 */
	@Override
	public void rootBoundsChanged(Rectangle rootBounds) {
		super.rootBoundsChanged(rootBounds);
		
		// template method for subclasses to perform rootbounds activities? maybe just subclass calling super is good enough here.
		rootBoundsChangedBeforeFirstPage(rootBounds);
		
		if (isResolutionSupported()) {
			if (waitingToInit) {
				waitingToInit = false;
				showFirstPage();
			}
		}
	}

	/**
	 * Resolution control is supported in protocol version 0.43 and SDK 1.4.1
	 */
	public boolean isResolutionSupported() {
		return this.protocolVersion >= VERSION_0_43 && isSDK141();
	}

	/**
	 * Stream Map argument is supported in protocol version 0.45 and SDK 1.4.1
	 */
	public boolean isStreamMapSupported() {
		return this.protocolVersion >= VERSION_0_45 && isSDK141();
	}

	/**
	 * Slide Remote is supported in protocol version 0.49
	 */
	public boolean isSlideRemoteSupported() {
		return this.protocolVersion >= VERSION_0_49;
	}

	/**
	 * Tries to get the version string from the com.tivo.hme.sdk.Version class
	 * which doesn't exist in all SDK versions. Returns null if it can't find
	 * the class.
	 * 
	 * @return
	 */
	public static String getSDKVersion() {
		try {
			if (Class.forName("com.tivo.hme.sdk.Version") != null) //$NON-NLS-1$
				return Version.getVersion();
			else
				return null;
		} catch (ClassNotFoundException e) {
			return null;
		}
	}

	public static boolean isSDK141() {
		if (getSDKVersion() != null)
			return getSDKVersion().startsWith("1.4.1"); //$NON-NLS-1$
		else
			return false;
	}

	/** a multi-SDK version of the SDK 1.4.1 method */
	public int getSafeActionHorizontal() {
		if (!isSDK141()) {
			return SAFE_ACTION_H;
		} else {
			return super.getSafeActionHorizontal();
		}
	}

	/** a multi-SDK version of the SDK 1.4.1 method */
	public int getSafeTitleVertical() {
		if (!isSDK141()) {
			return SAFE_TITLE_V;
		} else {
			return super.getSafeTitleVertical();
		}
	}

	/** a multi-SDK version of the SDK 1.4.1 method */
	public int getSafeActionVertical() {
		if (!isSDK141()) {
			return SAFE_ACTION_V;
		} else {
			return super.getSafeActionVertical();
		}
	}

	/** a multi-SDK version of the SDK 1.4.1 method */
	public int getSafeTitleHorizontal() {
		if (!isSDK141()) {
			return SAFE_TITLE_H;
		} else {
			return super.getSafeTitleHorizontal();
		}
	}

	
	/** Standard title area for e.g. a STREAM_LOOP_* background. */
	public Rectangle getTitleRegion() {
		int topPad = getSafeTitleVertical();
		int height = getSafeTitleToScrollTop()
				- 15;
		int leftPad = getSafeTitleHorizontal() + getSafeLogoToScrollTitle();
		int rightPad = getSafeTitleHorizontal();
		return new Rectangle(leftPad, topPad, getWidth()-leftPad-rightPad, height);
	}
	
	/** Standard logo area for e.g. a STREAM_LOOP_* background. */
	public Rectangle getLogoRegion() {
		int topPad = getSafeTitleVertical();
		int height = getSafeTitleToScrollTop()
				- 15;
		int leftPad = getSafeTitleHorizontal()
				- 20;
		int width = getSafeLogoToScrollTitle() + 10; // seems to fit, but shouldn't it be +20 to offset leftpad adjustment?
		return new Rectangle(leftPad, topPad, width, height);
	}

	/** Standard scroll area for e.g. a STREAM_LOOP_* background. */
	public Rectangle getScrollRegion() {
		int bottomPad = getSafeTitleVertical()+getSafeTitleToScrollBottom();
		int topPad = getSafeTitleVertical()+getSafeTitleToScrollTop();
		int leftPad = getSafeTitleHorizontal();
		int rightPad = getSafeTitleHorizontal();
		return new Rectangle(leftPad, topPad, getWidth()-leftPad-rightPad, getHeight()-topPad-bottomPad);
	}
	
	/**
	 * Vertical Space from safeTitle bottom to bottom of scrolling area of of a
	 * STREAM_LOOP_* background.
	 */
	public int getSafeTitleToScrollBottom() {
		if (isResolutionSupported() && getCurrentResolution().getHeight() > 480) {
			return 30;
		} else {
			return 15; // SD number
		}
	}

	/**
	 * Vertical Space from safeTitle top to top of scrolling area of a
	 * STREAM_LOOP_* background.
	 */
	public int getSafeTitleToScrollTop() {
		if (isResolutionSupported() && getCurrentResolution().getHeight() > 480) {
			return 85 + 60;
		} else {
			return 95; // SD number
		}
	}
	
	/**
	 * Horizontal space from safeTitle left to right edge of TiVo logo in the
	 * title area of a STREAM_LOOP_* background.  Note, logo extends a little to the left of the safeTitle left.
	 */
	public int getSafeLogoToScrollTitle() {
		if (isResolutionSupported() && getCurrentResolution().getHeight() > 480) {
			return 100; // could need a diff number for 720, but this will work.
		} else {
			return 45;// SD number - goes past "glow"
		}
	}
	
	protected void setMaxStackDepth(int max) {
		if (isSDK141()) {
			super.setMaxStackDepth(max);
		}
	}

	/**
	 * If {@link #isResolutionSupported()}, makes sure the resolution is the
	 * preferred one, and if it's not, calls
	 * {@link #setReceiverResolution(com.tivo.hme.sdk.Resolution)} with the
	 * ResolutionInfo's preferred resolution.
	 * 
	 * @return true if {@link #setReceiverResolution(com.tivo.hme.sdk.Resolution)} was actually called.
	 */
	protected boolean setPreferredResolution() {
			/* BEGIN HME 1.4.1e only */
			if (isResolutionSupported()) {
				boolean noChangeNeeded = this
						.getReceiverResolutionInfo()
						.getPreferredRenderingResolution()
						.equals(this.getReceiverResolutionInfo()
								.getRenderingResolution());
				if (noChangeNeeded) {
	//				debugOut(this.getCurrentResolution()+" is already preferred resolution of "+this.getReceiverResolutionInfo().getAvailableRenderingResolutions());
					return false;
				} else {
	//				 debugOut(this.getCurrentResolution()
	//						 +" => "+this.getReceiverResolutionInfo().getPreferredRenderingResolution()+" of "+this.getReceiverResolutionInfo().getAvailableRenderingResolutions());
					this.setReceiverResolution(this.getReceiverResolutionInfo()
							.getPreferredRenderingResolution());
					return true;
				}
			}
			/* END HME 1.4.1e only */
			return false;
		}

}