package com.blackledge.david.tivo;

import java.io.IOException;

import com.tivo.hme.interfaces.IApplication;
import com.tivo.hme.interfaces.IContext;
import com.tivo.hme.interfaces.ILogger;
import com.tivo.hme.sdk.Factory;
import com.tivo.hme.sdk.IHmeProtocol;
import com.tivo.hme.sdk.io.HmeInputStream;
import com.tivo.hme.sdk.io.HmeOutputStream;

/**
 * Factory that creates the {@link V49BApplication} to report its version to the
 * TiVo as VERSION_0_49. (V49BApplication required for createApplication
 * duplicate to have correct method access).
 */
public class V49BApplicationFactory extends Factory {
	/**
	 * Override to report a different HME version by calling
	 * {@link #supercreateApplication(IContext, int)}.
	 */
	@Override
	public IApplication createApplication(IContext context)
			throws IOException {
        int sentVersion = 
//                IHmeProtocol.VERSION;
        		V49IHmeProtocol.VERSION_0_49;
//        		IHmeProtocol.VERSION_0_40;
        
		IApplication application = supercreateApplication(context, sentVersion);
		
		return application;
	}
	
	/**
	 * Copy of super factory implementation with parameterized version number to
	 * send to TiVo.
	 */
    public IApplication supercreateApplication(IContext context, int sentVersion) throws IOException
    {
	int version = -1;
        HmeOutputStream out = null;
        HmeInputStream in = null;

        synchronized (lock) {                   // lock this factory
            if (context.getOutputStream() instanceof HmeOutputStream) {
                out = (HmeOutputStream)context.getOutputStream();
	    }
            else {
                out = new HmeOutputStream(context.getOutputStream());
	    }
            
            if (context.getInputStream() instanceof HmeInputStream) {
                in = (HmeInputStream)context.getInputStream();
            } else {
                in = new HmeInputStream(context.getInputStream());
	    }
            
            // HME protocol starts here
	    // Synchronize and flush the writes, so that no other writer 
	    // interleaves within our data
	    synchronized (out.getMonitor()) {
                out.writeInt(IHmeProtocol.MAGIC);
                out.writeInt(
//                		IHmeProtocol.VERSION
                		sentVersion
                		);
                out.flush();
	    }
		
            // read magic and version
            int magic = in.readInt();
            if (magic != IHmeProtocol.MAGIC) {
                throw new IOException(
		    "bad magic: 0x" + Integer.toHexString(magic)
//		    +getIdForException());
		    );
            }
            version = in.readInt();
            if (version >> 8 < 
//            IHmeProtocol.VERSION
            sentVersion
            >> 8 ) {
                throw new IOException(
                        "version mismatch: " + 
                        ( version >> 8 ) + "." + ( version & 0xff ) +
                        " < " +
                        ( 
//                        		IHmeProtocol.VERSION
                        		sentVersion
                        		>> 8 ) + "." + ( 
//                        				IHmeProtocol.VERSION
                        				sentVersion
                        				& 0xff )
//			+ getIdForException());
                        );
            }
        } // synchronized

        // maintain proper order of locks -- we have released 
        // the factory monitor before entering the app monitor
    
        IApplication retApp = null;
        try {
//            Application app = (Application)clazz.newInstance();
        	// have to have local class to access setContext
        	V49BApplication app = (V49BApplication)clazz.newInstance();
            synchronized (app) {
                app.setFactory(this);
                app.setContext(context, version);
		app.setMonitor(app);
                retApp = app;
            }
        } catch (InstantiationException ex) {
	    log(ILogger.LOG_NOTICE, ex);
        } catch (IllegalAccessException ex) {
	    log(ILogger.LOG_NOTICE, ex);
        }

	out.flush();

        return retApp;
    } // createApplication()
	
}