package com.blackledge.david.tivo;

import com.tivo.hme.bananas.BView;
import com.tivo.hme.sdk.HmeEvent;
import com.tivo.hme.sdk.HmeEvent.ResourceInfo;
import com.tivo.hme.sdk.StreamResource;

/**
 * This view just loads a video resource, and jumps it back to a starting point
 * every time it passes the end point. This is just a view...if you want to
 * scale it, crop it, or position it, that's up to you. Scaling doesn't work if
 * the position is against the top left corner. Positioning acts funny if it is
 * "moved off screen" - it actually crops after the visible video stops at the
 * edge of the screen You can specify a backup resource (an image or color
 * specification) - It is displayed while the video's starting point hasn't been
 * reached, if an error occurs in the video, or if you call closeVideo. If the
 * Video gets stuck at the end of the buffer downloading, it is supposed to
 * paused until the download completes, then continue automatically. NOTE some
 * video formats are only sensitive to within a few seconds of a playback
 * position, so the positions you give work exactly as you specified.
 */
public class VideoBackgroundView extends BView {

	private String backupResource;
	private int start;
	private int loopEnd;
	private int loopRestart;
	private String resourceSpec;
	private String resourceType;
	private StreamResource videoResource;
	private BView backupView;
	private boolean restarting;

	public VideoBackgroundView(BView parent) {
		super(parent, 0, 0, parent.getWidth(), parent.getHeight());
	}

	/**
	 * If loopEnd is not specified, it will loop when the video gets to its end.
	 * This is takes longer to recognize, making the loop less fluid, and can
	 * break, so don't do it. Figure out a decent value instead. ResourceType
	 * must match the resource or it might break. If start is greater than 0,
	 * backupResource shows in front of the video until start (or loopRestart)
	 * is passed. Attempts to jump to the start as soon as it's available in the
	 * buffer.
	 */
	public VideoBackgroundView(BView parent, String resourceSpec, String resourceType, int start,
			int loopEnd, int loopRestart, String backupResource) {
		this(parent);
		setVideo(resourceSpec, resourceType, start, loopEnd, loopRestart);
		setBackupResource(backupResource);
		startVideo();
	}

	public VideoBackgroundView(BView parent, String resourceSpec, String resourceType, int start,
			int loopEnd, int loopRestart) {
		this(parent, resourceSpec, resourceType, start, loopEnd, loopRestart, null);
	}

	public void setVideo(String resourceSpec, String resourceType, int start, int loopEnd,
			int loopRestart) {
		this.start = start;
		this.loopEnd = loopEnd;
		this.loopRestart = loopRestart;
		this.resourceSpec = resourceSpec;
		this.resourceType = resourceType;
		if (start < 0)
			this.start = 0;
		if (loopRestart < 0)
			this.loopRestart = start;
		
		
		if (this.getResource() != null) {
		}
	}
	
	public void setVideoNowPlaying() {
		setVideo(V49BApplication.STREAM_LOOP_GREEN_NOWPLAYING, "video/mpeg", -1, -1, -1);
	}
	public void setVideoSetup() {
		setVideo(V49BApplication.STREAM_LOOP_BLUE_SETUP, "video/mpeg", -1, -1, -1);
	}
	public void setVideoShowcases() {
		setVideo(V49BApplication.STREAM_LOOP_PURPLE_SHOWCASES, "video/mpeg", -1, -1, -1);
	}
	public void setVideoCentral() {
		setVideo(V49BApplication.STREAM_LOOP_RED_CENTRAL, "video/mpeg", -1, -1, -1);
	}

	public void setBackupResource(String resourceSpec) {
		this.backupResource = resourceSpec;
		if (this.getResource() == null)
			showBackupResource();
	}

	public void startVideo() {
		if (this.getResource() != null)
			return;
		this.videoResource = createStream(resourceSpec, resourceType, true);
		// loopsets are self-looping
		if (!resourceSpec.startsWith("loopset:")) {
			this.videoResource.addHandler(this);
		}
		this.setResource(this.videoResource);
	}
	
	@Override
	public boolean handleEvent(HmeEvent event) {
		//ignore if it's not our resource... 
		//probably switched videos and we have leftover events.
		if(event instanceof ResourceInfo) {
			ResourceInfo revent = (ResourceInfo)event;
			if(this.videoResource != null && revent.getResource().getID() == this.videoResource.getID()) {
				int state = revent.getStatus();
				if(state == V49IHmeProtocol.RSRC_STATUS_END) {
					if(! this.restarting && this.getResource() != null) {
						this.restarting = true;
						((StreamResource) this.getResource()).setPosition(loopRestart);
						((StreamResource) this.getResource()).setSpeed(1.0f);
					}
				} else if(state == V49IHmeProtocol.RSRC_STATUS_BUFFER_OVERFLOW) {
	
					if (!this.restarting && this.getResource() != null) {
						if (((StreamResource) this.getResource()).getSpeed() != 1.0f) {
							((StreamResource) this.getResource()).setSpeed(1.0f);
						}
					}
				} else if (state == RSRC_STATUS_CLOSED
						|| state == RSRC_STATUS_CONNECTED
						|| state == RSRC_STATUS_CONNECTING
						|| state == RSRC_STATUS_LOADING
						|| state == RSRC_STATUS_READY
						|| state == RSRC_STATUS_SEEKING
						|| state == RSRC_STATUS_UNKNOWN) {
					// not playing
					showBackupResource();
				} else {
					// complete, playing, paused
					int pos = getPosition(revent);
					if (pos >= 0) {
						if (pos < start && pos < loopRestart) {
							// this.setVisible(false);
							showBackupResource();
							// attempt to setPos to start?
							int dur = getDuration(revent);
							if(dur >= start) {
								((StreamResource) this.getResource()).setPosition(start);
							}
						} else {
							// this.setVisible(true);
							hideBackupResource();
							if (loopEnd > -1 && pos >= loopEnd) {
								if (!this.restarting && this.getResource() != null) {
									this.restarting = true;
									((StreamResource) this.getResource()).setPosition(loopRestart);
								}
							} else {
								if (this.restarting) {
									this.restarting = false;
								}
							}
						}
					}
				}
				return true;
			}
		}
		return super.handleEvent(event);
	}

	private int getPosition(ResourceInfo event) {
		try {
			return Integer.parseInt(String.valueOf(event.getMap().get(
					"position")));
		} catch (Exception e) {
			return -1;
		}
	}

	private int getDuration(ResourceInfo event) {
		try {
			return Integer.parseInt(String.valueOf(event.getMap().get(
					"duration")));
		} catch (Exception e) {
			return -1;
		}
	}

	public void closeVideo() {
		showBackupResource();
		if (this.getResource() != null) {
			// don't actually remove it so startVideo works quickly.
			// this.getResource().remove();
			this.clearResource();
		}
	}

	private void showBackupResource() {
		if (this.backupResource != null) {
			// backup resource shouldn't be used when we have a loopset
			// since it won't be closed again.
			if (this.resourceSpec != null
					&& this.resourceSpec.startsWith("loopset:")) {
				return;
			}

			if (this.backupView == null) {
				this.backupView = new BView(this, 0, 0, this.getWidth(),
						this.getHeight(), false);
			}
			this.backupView.setResource(this.backupView
					.getResource(this.backupResource),
					RSRC_HALIGN_CENTER | RSRC_VALIGN_CENTER
					| RSRC_IMAGE_BESTFIT);
			this.backupView.setVisible(true);
		}
	}

	private void hideBackupResource() {
		if (this.backupView != null) {
			this.backupView.setVisible(false);
		}
	}

}
